// To compile: gcc -O3 -o nh File-Name rijndael-alg-fst.h rijndael-alg-fst.c

#include <string.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <time.h>
#include "rijndael-alg-fst.h"

typedef unsigned char      UINT8;  /* 1 byte   */
typedef unsigned short     UINT16; /* 2 byte   */
typedef unsigned int       UINT32; /* 4 byte   */
typedef unsigned long long UINT64; /* 8 bytes  */
typedef unsigned long      UWORD;  /* Register */

#define NH_KEY_LEN           16  /* NH takes 16 bytes of external key */
#define MUL64(a,b) ((UINT64)((UINT64)(UINT32)(a) * (UINT64)(UINT32)(b)))
#define LOAD_UINT32_LITTLE(ptr)     (*(UINT32 *)(ptr))



#define NH_OUTPUT_LEN       64       /* Allowable: 8,16,24,32,40,64 bytes */
#define STREAMS (NH_OUTPUT_LEN / 8) /* Number of times hash is applied  */
#define L1_KEY_LEN         1024       /* Internal key bytes                 */
#define L1_KEY_SHIFT         16       /* Toeplitz key shift between streams */

/* ---------------------------------------------------------------------- */
/* ----- Begin KDF & PDF Section ---------------------------------------- */
/* ---------------------------------------------------------------------- */

/* NH uses AES with 16 byte block and key lengths */
#define AES_BLOCK_LEN  16
#define AES_ROUNDS	   ((NH_KEY_LEN / 4) + 6)
typedef UINT8          aes_int_key[AES_ROUNDS+1][4][4]; /* AES internal */
#define aes_encryption(in,out,int_key) \
	    rijndaelEncrypt((u32 *)(int_key), AES_ROUNDS, (u8 *)(in), (u8 *)(out))
#define aes_key_setup(key,int_key) \
	    rijndaelKeySetupEnc((u32 *)(int_key), (const unsigned char *)(key), \
	    NH_KEY_LEN*8)


/* The user-supplied NH key is stretched using AES in a counter
 * mode to supply all random bits needed by NH. The kdf function takes
 * an AES internal key representation 'key' and writes a stream of
 * 'nbytes' bytes to the memory pointed at by 'buffer_ptr'. Each distinct
 * 'index' causes a distinct byte stream.
 */
void kdf(void *buffer_ptr, aes_int_key key, UINT8 index, int nbytes)
{
    UINT8 in_buf[AES_BLOCK_LEN] = {0};
    UINT8 out_buf[AES_BLOCK_LEN];
    UINT8 *dst_buf = (UINT8 *)buffer_ptr;
    int i;
    
    /* Setup the initial value */
    in_buf[AES_BLOCK_LEN-9] = index;
    in_buf[AES_BLOCK_LEN-1] = i = 1;
        
    while (nbytes >= AES_BLOCK_LEN) {
        aes_encryption(in_buf, out_buf, key);
        memcpy(dst_buf,out_buf,AES_BLOCK_LEN);
        in_buf[AES_BLOCK_LEN-1] = ++i;
        nbytes -= AES_BLOCK_LEN;
        dst_buf += AES_BLOCK_LEN;
    }
    if (nbytes) {
        aes_encryption(in_buf, out_buf, key);
        memcpy(dst_buf,out_buf,nbytes);
    }
}

#if (NH_OUTPUT_LEN == 8) // TWO 32-BIT WORDS OR 64 BITS

static void nh_aux(void *kp, void *dp, void *hp, UINT32 dlen)
{
    UINT64 h;
    UWORD c = dlen / 32;
    UINT32 *k = (UINT32 *)kp;
    UINT32 *d = (UINT32 *)dp;
    UINT32 d0,d1,d2,d3,d4,d5,d6,d7;
    UINT32 k0,k1,k2,k3,k4,k5,k6,k7,k8;
    
    h = *((UINT64 *)hp);
    do {
        d0 = LOAD_UINT32_LITTLE(d+0); d1 = LOAD_UINT32_LITTLE(d+1);
        d2 = LOAD_UINT32_LITTLE(d+2); d3 = LOAD_UINT32_LITTLE(d+3);
        d4 = LOAD_UINT32_LITTLE(d+4); d5 = LOAD_UINT32_LITTLE(d+5);
        d6 = LOAD_UINT32_LITTLE(d+6); d7 = LOAD_UINT32_LITTLE(d+7);

        k0 = *(k+0); k1 = *(k+1); k2 = *(k+2); k3 = *(k+3);
        k4 = *(k+4); k5 = *(k+5); k6 = *(k+6); k7 = *(k+7);

        h += MUL64((k0 + d0), (k1 + d1));
        h += MUL64((k2 + d2), (k3 + d3));
        h += MUL64((k4 + d4), (k5 + d5));
        h += MUL64((k6 + d6), (k7 + d7));

        d += 8;
        k += 8;
    } while (--c);
  *((UINT64 *)hp) = h;
}

#elif (NH_OUTPUT_LEN == 16) // FOUR 32-BIT WORDS OR 128 BITS

static void nh_aux(void *kp, void *dp, void *hp, UINT32 dlen)
/* Same as previous nh_aux, but two streams are handled in one pass,
 * reading and writing 16 bytes of hash-state per call.
 */
{
  UINT64 h1,h2;
  UWORD c = dlen / 32;
  UINT32 *k = (UINT32 *)kp;
  UINT32 *d = (UINT32 *)dp;
  UINT32 d0,d1,d2,d3,d4,d5,d6,d7;
  UINT32 k0,k1,k2,k3,k4,k5,k6,k7,
        k8,k9,k10,k11;

  h1 = *((UINT64 *)hp);
  h2 = *((UINT64 *)hp + 1);
  k0 = *(k+0); k1 = *(k+1); k2 = *(k+2); k3 = *(k+3);
  do {
    d0 = LOAD_UINT32_LITTLE(d+0); d1 = LOAD_UINT32_LITTLE(d+1);
    d2 = LOAD_UINT32_LITTLE(d+2); d3 = LOAD_UINT32_LITTLE(d+3);
    d4 = LOAD_UINT32_LITTLE(d+4); d5 = LOAD_UINT32_LITTLE(d+5);
    d6 = LOAD_UINT32_LITTLE(d+6); d7 = LOAD_UINT32_LITTLE(d+7);
    k4 = *(k+4); k5 = *(k+5); k6 = *(k+6); k7 = *(k+7);
    k8 = *(k+8); k9 = *(k+9); k10 = *(k+10); k11 = *(k+11);

    h1 += MUL64((k0 + d0), (k4 + d4));
    h2 += MUL64((k4 + d0), (k8 + d4));

    h1 += MUL64((k1 + d1), (k5 + d5));
    h2 += MUL64((k5 + d1), (k9 + d5));

    h1 += MUL64((k2 + d2), (k6 + d6));
    h2 += MUL64((k6 + d2), (k10 + d6));

    h1 += MUL64((k3 + d3), (k7 + d7));
    h2 += MUL64((k7 + d3), (k11 + d7));

    k0 = k8; k1 = k9; k2 = k10; k3 = k11;

    d += 8;
    k += 8;
  } while (--c);
  ((UINT64 *)hp)[0] = h1;
  ((UINT64 *)hp)[1] = h2;
}

#elif (NH_OUTPUT_LEN == 24) // SIX 32-BIT WORDS OR 192 BITS

static void nh_aux(void *kp, void *dp, void *hp, UINT32 dlen)
/* Same as previous nh_aux, but two streams are handled in one pass,
 * reading and writing 24 bytes of hash-state per call.
*/
{
    UINT64 h1,h2,h3;
    UWORD c = dlen / 32;
    UINT32 *k = (UINT32 *)kp;
    UINT32 *d = (UINT32 *)dp;
    UINT32 d0,d1,d2,d3,d4,d5,d6,d7;
    UINT32 k0,k1,k2,k3,k4,k5,k6,k7,
        k8,k9,k10,k11,k12,k13,k14,k15;
    
    h1 = *((UINT64 *)hp);
    h2 = *((UINT64 *)hp + 1);
    h3 = *((UINT64 *)hp + 2);
    k0 = *(k+0); k1 = *(k+1); k2 = *(k+2); k3 = *(k+3);
    k4 = *(k+4); k5 = *(k+5); k6 = *(k+6); k7 = *(k+7);
    do {
        d0 = LOAD_UINT32_LITTLE(d+0); d1 = LOAD_UINT32_LITTLE(d+1);
        d2 = LOAD_UINT32_LITTLE(d+2); d3 = LOAD_UINT32_LITTLE(d+3);
        d4 = LOAD_UINT32_LITTLE(d+4); d5 = LOAD_UINT32_LITTLE(d+5);
        d6 = LOAD_UINT32_LITTLE(d+6); d7 = LOAD_UINT32_LITTLE(d+7);
        k8 = *(k+8); k9 = *(k+9); k10 = *(k+10); k11 = *(k+11);
        k12 = *(k+12); k13 = *(k+13); k14 = *(k+14); k15 = *(k+15);
        
        h1 += MUL64((k0 + d0), (k4 + d4));
        h2 += MUL64((k4 + d0), (k8 + d4));
        h3 += MUL64((k8 + d0), (k12 + d4));
        
        h1 += MUL64((k1 + d1), (k5 + d5));
        h2 += MUL64((k5 + d1), (k9 + d5));
        h3 += MUL64((k9 + d1), (k13 + d5));
        
        h1 += MUL64((k2 + d2), (k6 + d6));
        h2 += MUL64((k6 + d2), (k10 + d6));
        h3 += MUL64((k10 + d2), (k14 + d6));
        
        h1 += MUL64((k3 + d3), (k7 + d7));
        h2 += MUL64((k7 + d3), (k11 + d7));
        h3 += MUL64((k11 + d3), (k15 + d7));
        
        k0 = k8; k1 = k9; k2 = k10; k3 = k11;
        k4 = k12; k5 = k13; k6 = k14; k7 = k15;
        
        d += 8;
        k += 8;
    } while (--c);
    ((UINT64 *)hp)[0] = h1;
    ((UINT64 *)hp)[1] = h2;
    ((UINT64 *)hp)[2] = h3;
}

#elif (NH_OUTPUT_LEN == 32) // EIGHT 32-BIT WORDS OR 256 BITS

static void nh_aux(void *kp, void *dp, void *hp, UINT32 dlen)
/* Same as previous nh_aux, but two streams are handled in one pass,
 * reading and writing 24 bytes of hash-state per call.
*/
{
    UINT64 h1,h2,h3,h4;
    UWORD c = dlen / 32;
    UINT32 *k = (UINT32 *)kp;
    UINT32 *d = (UINT32 *)dp;
    UINT32 d0,d1,d2,d3,d4,d5,d6,d7;
    UINT32 k0,k1,k2,k3,k4,k5,k6,k7,
        k8,k9,k10,k11,k12,k13,k14,k15,
        k16,k17,k18,k19;
    
    h1 = *((UINT64 *)hp);
    h2 = *((UINT64 *)hp + 1);
    h3 = *((UINT64 *)hp + 2);
    h4 = *((UINT64 *)hp + 3);

    k0 = *(k+0); k1 = *(k+1); k2 = *(k+2); k3 = *(k+3);
    k4 = *(k+4); k5 = *(k+5); k6 = *(k+6); k7 = *(k+7);
    k8 = *(k+8); k9 = *(k+9); k10 = *(k+10); k11 = *(k+11);
    do {
        d0 = LOAD_UINT32_LITTLE(d+0); d1 = LOAD_UINT32_LITTLE(d+1);
        d2 = LOAD_UINT32_LITTLE(d+2); d3 = LOAD_UINT32_LITTLE(d+3);
        d4 = LOAD_UINT32_LITTLE(d+4); d5 = LOAD_UINT32_LITTLE(d+5);
        d6 = LOAD_UINT32_LITTLE(d+6); d7 = LOAD_UINT32_LITTLE(d+7);
        k12 = *(k+12); k13 = *(k+13); k14 = *(k+14); k15 = *(k+15);
        k16 = *(k+16); k17 = *(k+17); k18 = *(k+18); k19 = *(k+19);
        
        h1 += MUL64((k0 + d0), (k4 + d4));
        h2 += MUL64((k4 + d0), (k8 + d4));
        h3 += MUL64((k8 + d0), (k12 + d4));
        h4 += MUL64((k12 + d0), (k16 + d4));
        
        h1 += MUL64((k1 + d1), (k5 + d5));
        h2 += MUL64((k5 + d1), (k9 + d5));
        h3 += MUL64((k9 + d1), (k13 + d5));
        h4 += MUL64((k13 + d1), (k17 + d5));
        
        h1 += MUL64((k2 + d2), (k6 + d6));
        h2 += MUL64((k6 + d2), (k10 + d6));
        h3 += MUL64((k10 + d2), (k14 + d6));
        h4 += MUL64((k14 + d2), (k18 + d6));
        
        h1 += MUL64((k3 + d3), (k7 + d7));
        h2 += MUL64((k7 + d3), (k11 + d7));
        h3 += MUL64((k11 + d3), (k15 + d7));
        h4 += MUL64((k15 + d3), (k19 + d7));
        
        k0 = k8; k1 = k9; k2 = k10; k3 = k11;
        k4 = k12; k5 = k13; k6 = k14; k7 = k15;
        k8 = k16; k9 = k17; k10 = k18; k11 = k19;
        
        d += 8;
        k += 8;
    } while (--c);
    ((UINT64 *)hp)[0] = h1;
    ((UINT64 *)hp)[1] = h2;
    ((UINT64 *)hp)[2] = h3;
    ((UINT64 *)hp)[3] = h4;
}

#elif (NH_OUTPUT_LEN == 40) // TEN 32-BIT WORDS OR 320 BITS

static void nh_aux(void *kp, void *dp, void *hp, UINT32 dlen)
{
    UINT64 h1,h2,h3,h4,h5;
    UWORD c = dlen / 32;
    UINT32 *k = (UINT32 *)kp;
    UINT32 *d = (UINT32 *)dp;
    UINT32 d0,d1,d2,d3,d4,d5,d6,d7;
    UINT32 k0,k1,k2,k3,k4,k5,k6,k7,
        k8,k9,k10,k11,k12,k13,k14,k15,
      k16,k17,k18,k19,k20,k21,k22,k23;
    
    h1 = *((UINT64 *)hp);
    h2 = *((UINT64 *)hp + 1);
    h3 = *((UINT64 *)hp + 2);
    h4 = *((UINT64 *)hp + 3);
    h5 = *((UINT64 *)hp + 4);

    k0 = *(k+0); k1 = *(k+1); k2 = *(k+2); k3 = *(k+3);
    k4 = *(k+4); k5 = *(k+5); k6 = *(k+6); k7 = *(k+7);
    k8 = *(k+8); k9 = *(k+9); k10 = *(k+10); k11 = *(k+11);
    k12 = *(k+12); k13 = *(k+13); k14 = *(k+14); k15 = *(k+15);

    do {
        d0 = LOAD_UINT32_LITTLE(d+0); d1 = LOAD_UINT32_LITTLE(d+1);
        d2 = LOAD_UINT32_LITTLE(d+2); d3 = LOAD_UINT32_LITTLE(d+3);
        d4 = LOAD_UINT32_LITTLE(d+4); d5 = LOAD_UINT32_LITTLE(d+5);
        d6 = LOAD_UINT32_LITTLE(d+6); d7 = LOAD_UINT32_LITTLE(d+7);

        k16 = *(k+16); k17 = *(k+17); k18 = *(k+18); k19 = *(k+19);
        k20 = *(k+20); k21 = *(k+21); k22 = *(k+22); k23 = *(k+23);
        	
        h1 += MUL64((k0 + d0), (k4 + d4));
        h2 += MUL64((k4 + d0), (k8 + d4));
        h3 += MUL64((k8 + d0), (k12 + d4));
        h4 += MUL64((k12 + d0), (k16 + d4));
        h5 += MUL64((k16 + d0), (k20 + d4));
        
        h1 += MUL64((k1 + d1), (k5 + d5));
        h2 += MUL64((k5 + d1), (k9 + d5));
        h3 += MUL64((k9 + d1), (k13 + d5));
        h4 += MUL64((k13 + d1), (k17 + d5));
        h5 += MUL64((k17 + d1), (k21 + d5));        

        h1 += MUL64((k2 + d2), (k6 + d6));
        h2 += MUL64((k6 + d2), (k10 + d6));
        h3 += MUL64((k10 + d2), (k14 + d6));
        h4 += MUL64((k14 + d2), (k18 + d6));
        h5 += MUL64((k18 + d2), (k22 + d6));        

        h1 += MUL64((k3 + d3), (k7 + d7));
        h2 += MUL64((k7 + d3), (k11 + d7));
        h3 += MUL64((k11 + d3), (k15 + d7));
        h4 += MUL64((k15 + d3), (k19 + d7));
        h5 += MUL64((k19 + d3), (k23 + d7));        

        k0 = k8; k1 = k9; k2 = k10; k3 = k11;
        k4 = k12; k5 = k13; k6 = k14; k7 = k15;
	k8 = k16; k9 = k17; k10 = k18; k11 = k19;
	k12 = k20; k13 = k21; k14 = k22; k15 = k23;

        d += 8;
        k += 8;
    } while (--c);
    ((UINT64 *)hp)[0] = h1;
    ((UINT64 *)hp)[1] = h2;
    ((UINT64 *)hp)[2] = h3;
    ((UINT64 *)hp)[3] = h4;
    ((UINT64 *)hp)[4] = h5;
}

#elif (NH_OUTPUT_LEN == 64) // SIXTEEN 32-BIT WORDS, OR 512 BITS

static void nh_aux(void *kp, void *dp, void *hp, UINT32 dlen)
{
    UINT64 h1,h2,h3,h4,h5,h6,h7,h8;
    UINT64 m1,m2,m3,m4,m5,m6,m7;
    UWORD c = dlen / 32;
    UINT32 *k = (UINT32 *)kp;
    UINT32 *d = (UINT32 *)dp;
    UINT32 d0,d1,d2,d3,d4,d5,d6,d7;
    UINT32 k0,k1,k2,k3,k4,k5,k6,k7,
      k8,k9,k10,k11,k12,k13,k14,k15,
      k16,k17,k18,k19,k20,k21,k22,k23,
      k24,k25,k26,k27,k28,k29,k30,k31,
      k32,k33,k34,k35;
    
    h1 = *((UINT64 *)hp);
    h2 = *((UINT64 *)hp + 1);
    h3 = *((UINT64 *)hp + 2);
    h4 = *((UINT64 *)hp + 3);
    h5 = *((UINT64 *)hp + 4);
    h6 = *((UINT64 *)hp + 5);
    h7 = *((UINT64 *)hp + 6);
    h8 = *((UINT64 *)hp + 7);

    k0 = *(k+0); k1 = *(k+1); k2 = *(k+2); k3 = *(k+3);
    k4 = *(k+4); k5 = *(k+5); k6 = *(k+6); k7 = *(k+7);
    k8 = *(k+8); k9 = *(k+9); k10 = *(k+10); k11 = *(k+11);
    k12 = *(k+12); k13 = *(k+13); k14 = *(k+14); k15 = *(k+15);
    k16 = *(k+16) ; k17 = *(k+17) ; k18 = *(k+18) ; k19 = *(k+19) ;
    k20 = *(k+20) ; k21 = *(k+21) ; k22 = *(k+22) ; k23 = *(k+23) ;
    k24 = *(k+24) ; k25 = *(k+25) ; k26 = *(k+26) ; k27 = *(k+27) ;

    do {
        d0 = LOAD_UINT32_LITTLE(d+0); d1 = LOAD_UINT32_LITTLE(d+1);
        d2 = LOAD_UINT32_LITTLE(d+2); d3 = LOAD_UINT32_LITTLE(d+3);
        d4 = LOAD_UINT32_LITTLE(d+4); d5 = LOAD_UINT32_LITTLE(d+5);
        d6 = LOAD_UINT32_LITTLE(d+6); d7 = LOAD_UINT32_LITTLE(d+7);

        k28 = *(k+24); k29 = *(k+25); k30 = *(k+26); k31 = *(k+27);
        k32 = *(k+20); k33 = *(k+21); k34 = *(k+22); k35 = *(k+23);

        h1 += MUL64((k0 + d0), (k4 + d4));
        h2 += MUL64((k4 + d0), (k8 + d4));
        h3 += MUL64((k8 + d0), (k12 + d4));
        h4 += MUL64((k12 + d0), (k16 + d4));
        h5 += MUL64((k16 + d0), (k20 + d4));
        h6 += MUL64((k20 + d0), (k24 + d4));
        h7 += MUL64((k24 + d0), (k28 + d4));
        h8 += MUL64((k28 + d0), (k32 + d4));
        
        h1 += MUL64((k1 + d1), (k5 + d5));
        h2 += MUL64((k5 + d1), (k9 + d5));
        h3 += MUL64((k9 + d1), (k13 + d5));
        h4 += MUL64((k13 + d1), (k17 + d5));
        h5 += MUL64((k17 + d1), (k21 + d5));        
        h6 += MUL64((k21 + d1), (k25 + d5)); 
        h7 += MUL64((k25 + d1), (k29 + d5));        
        h8 += MUL64((k29 + d1), (k33 + d5));        

        h1 += MUL64((k2 + d2), (k6 + d6));
        h2 += MUL64((k6 + d2), (k10 + d6));
        h3 += MUL64((k10 + d2), (k14 + d6));
        h4 += MUL64((k14 + d2), (k18 + d6));
        h5 += MUL64((k18 + d2), (k22 + d6));        
        h6 += MUL64((k22 + d2), (k26 + d6));        
        h7 += MUL64((k26 + d2), (k30 + d6));        
        h8 += MUL64((k30 + d2), (k34 + d6));        

        h1 += MUL64((k3 + d3), (k7 + d7));
        h2 += MUL64((k7 + d3), (k11 + d7));
        h3 += MUL64((k11 + d3), (k15 + d7));
        h4 += MUL64((k15 + d3), (k19 + d7));
        h5 += MUL64((k19 + d3), (k23 + d7));        
        h6 += MUL64((k23 + d3), (k27 + d7));        
        h7 += MUL64((k27 + d3), (k31 + d7));        
        h8 += MUL64((k31 + d3), (k35 + d7));        

        k0 = k8; k1 = k9; k2 = k10; k3 = k11;
        k4 = k12; k5 = k13; k6 = k14; k7 = k15;
	k8 = k16; k9 = k17; k10 = k18; k11 = k19;
	k12 = k20; k13 = k21; k14 = k22; k15 = k23;
	k16 = k24; k17 = k25; k18 = k26; k19 = k27;
	k20 = k28; k21 = k29; k22 = k30; k23 = k31;
	k24 = k32; k25 = k33; k26 = k34; k27 = k35;
	
        d += 8;
        k += 8;
	
	
    } while (--c);
    ((UINT64 *)hp)[0] = h1;
    ((UINT64 *)hp)[1] = h2;
    ((UINT64 *)hp)[2] = h3;
    ((UINT64 *)hp)[3] = h4;
    ((UINT64 *)hp)[4] = h5;
    ((UINT64 *)hp)[5] = h6;
    ((UINT64 *)hp)[6] = h7;
    ((UINT64 *)hp)[7] = h8;
}

#endif


                                      


static double run_cpb_test(UINT8 nh_key[], int nbytes, char *data_ptr,
                           int data_len, double hz)
{
    clock_t ticks;
    double secs;
    unsigned long iters_per_tag;
    unsigned long tag_iters, i, no_iters;
    unsigned long total_mbs;

    int len;
    char *input;
    UINT8 nh_result[STREAMS*sizeof(UINT64)];

    total_mbs = 2500;
    tag_iters = (total_mbs * 1024 * 1024) / (nbytes) + 1;    
    iters_per_tag = nbytes / data_len;
    no_iters = iters_per_tag * tag_iters;

    //===================================================                
    len = data_len;
    input = data_ptr;
    while (len >= L1_KEY_LEN) {
      nh_aux(nh_key, input, nh_result, L1_KEY_LEN);
      len -= L1_KEY_LEN;
      input += L1_KEY_LEN;
    }             
    //==================================================

    ticks = clock();
    for(i = 0; i < no_iters; i++){
	len = data_len;
	input = data_ptr;

	while (len >= L1_KEY_LEN) {
	  nh_aux(nh_key, input, nh_result, L1_KEY_LEN);
	  len -= L1_KEY_LEN;
	  input += L1_KEY_LEN;
	}
    }
    ticks = clock() - ticks;
    secs = (double)ticks / CLOCKS_PER_SEC;

    return (secs * (hz/(tag_iters * nbytes)));
}

int main(void)
{
    aes_int_key prf_key;
    UINT8  nh_key [L1_KEY_LEN + L1_KEY_SHIFT * (STREAMS - 1)];
    char *data_ptr;
    int data_len, i;
    double hz;
    double cpb;

    //int length_pts = 2048*1024; 
    int length_pts[] = {8*1024,32*1024,256*1024,1024*1024};
    
    hz = ((double)1e9);
    data_len = 8192;

    data_ptr = (char *)malloc(data_len + 16);

    for (i = 0; i < data_len; i++)
        data_ptr[i] = (i*i) % 128;

    aes_key_setup("abcdefghijklmnopqrstuvwxyz",prf_key);
    kdf(nh_key, prf_key, 1, sizeof(nh_key));
        
    //cpb = run_cpb_test(nh_key, length_pts, data_ptr, data_len, hz);
    //printf("\nAuthenticating %8d byte messages: %5.2f cpb.\n", length_pts, cpb);

    printf("\nSpeed of %d-bit output NH(k,m) \n",NH_OUTPUT_LEN * 8);
    if (sizeof(length_pts) > 0) {
        for (i = 0; i < (int)(sizeof(length_pts)/sizeof(int)); i++) {
            cpb = run_cpb_test(nh_key, length_pts[i], data_ptr, data_len, hz);
            printf("NH hashing %8d byte messages: %5.2f cpb.\n",
                                   length_pts[i], cpb);
        }
    }
    printf("\n");
}




